from mext.threads.locks import SimpleLock, TimeoutLock
from thread import get_ident

__all__ = ['Condition', 'Semaphore', 'Event', 'Future']


class Condition(object):
    def __init__(self, lock=None):
        self.lock = lock or SimpleLock()
        self.waiters = []

    def __enter__(self):
        return self.lock.__enter__()

    def __exit__(self, *args):
        return self.lock.__exit__(*args)

    def __repr__(self):
        return "<%s(%d)>" % (self.__class__.__name__, len(self.waiters))

    def _check_locked(self):
        assert self.lock.locked(), "cannot wait on un-acquired lock"

    def wait(self, timeout=None):
        self._check_locked()
        if not timeout: # None or 0 can be served by simple Lock
            waiter = SimpleLock()
        else:
            waiter = TimeoutLock()
        waiter.acquire()
        self.waiters.append(waiter)
        self.lock.release()
        try:
            if not timeout:
                waiter.acquire(timeout is None)
                return True
            else:
                r = waiter.acquire(timeout)
        finally:
            self.lock.acquire()
        if not r:
            try:
                self.waiters.remove(waiter)
            except ValueError:
                pass
        return r

    def notify(self, n=1):
        self._check_locked()
        for waiter in self.waiters[:n]:
            waiter.release()
        self.waiters[:n] = []

    def notify_all(self):
        self.notify(len(self.waiters))




class Semaphore(object):
    def __init__(self, value=1, verbose=None):
        if value < 0:
            raise ValueError("semaphore initial valuet must be >= 0")
        self.cond = Condition()
        self.value = value

    def acquire(self, blocking=True):
        with self.cond:
            while self.value == 0:
                if not blocking:
                    return False
                self.cond.wait()
            self.value -= 1
            return True

    __enter__ = acquire

    def release(self):
        with self.cond:
            self.value += 1
            self.cond.notify()

    def __exit__(self, t, v, tb):
        self.release()



class Event(object):
    def __init__(self):
        self.cond = Condition()
        self.flag = False

    def is_set(self):
        return self.flag

    def set(self):
        with self.cond:
            self.flag = True
            self.cond.notify_all()

    def clear(self):
        with self.cond:
            self.flag = False

    def wait(self, timeout=None):
        with self.cond:
            if not self.flag:
                self.cond.wait(timeout)


class Future(object):
    def __init__(self):
        self.cond = Condition()
        self.ready = False

    #def read(self, timeout=None):
    def read(self):
        with self.cond:
            if not self.ready:
                self.cond.wait()#timeout)
        return self.value

    def write(self, value):
        with self.cond:
            if self.ready:
                raise RuntimeError("Only write a future once")
            self.ready = True
            self.value = value
            self.cond.notify_all()

    #get, set = read, write
